from stash.codecs.codec import Codec

try:
    import brotli
except ImportError:
    pass


class BrotliCodec(Codec):
    def encode(self, data):
        return brotli.compress(data)

    def decode(self, data):
        return brotli.decompress(data)



from stash.serializers.serializer import Serializer
import pickle


class DefaultSerializer(Serializer):
    def deserialize(self, data):
        return pickle.loads(data)

    def serialize(self, data):
        return pickle.dumps(data)



from abc import ABC, abstractmethod

from stash.options import StashOptions


class Storage(ABC):
    def __init__(self, options: StashOptions):
        self.options = options

    @abstractmethod
    def exists(self, key: str) -> bool:
        return False

    @abstractmethod
    def purge(self, cutoff: int):
        pass

    @abstractmethod
    def clear(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def write(self, key: str, content):
        pass

    @abstractmethod
    def read(self, key: str):
        pass

    @abstractmethod
    def rm(self, key: str):
        pass



from stash.options import StashOptions
from stash.storages.storage import Storage


class MemoryStorage(Storage):
    def __init__(self, options: StashOptions):
        super().__init__(options)
        self.__dict = {}

    def exists(self, key: str) -> bool:
        return key in self.__dict

    def purge(self, cutoff: int):
        pass

    def clear(self):
        self.__dict.clear()

    def close(self):
        self.clear()

    def write(self, key: str, content):
        self.__dict[key] = content

    def read(self, key: str):
        return self.__dict.get(key)

    def rm(self, key: str):
        self.__dict.pop(key, None)



from stash.codecs.codec import Codec

try:
    import lzma
except ImportError:
    try:
        import backports.lzma as lzma
    except ImportError:
        pass


class LzmaCodec(Codec):
    def encode(self, data):
        return lzma.compress(data)

    def decode(self, data):
        return lzma.decompress(data)



from typing import List

import stash.consts
from stash._helpers import *
from stash.codecs.brotli import BrotliCodec
from stash.codecs.bzip2 import BZip2Codec
from stash.codecs.codec import Codec
from stash.codecs.gzip import GZipCodec
from stash.codecs.lz4 import Lz4Codec
from stash.codecs.lzf import LzfCodec
from stash.codecs.lzma import LzmaCodec
from stash.codecs.lzo import LzoCodec
from stash.codecs.passthru import PassthruCodec
from stash.codecs.snappy import SnappyCodec
from stash.codecs.zlib import ZlibCodec
from stash.codecs.zopfli import ZopfliCodec
from stash.codecs.zstd import ZstdCodec
from stash.manager import StashManager
from stash.serializers.bson import BSONSerializer
from stash.serializers.cbor import CBORSerializer
from stash.serializers.default import DefaultSerializer
from stash.serializers.json import JSONSerializer
from stash.serializers.msgpack import MsgPackSerializer
from stash.serializers.orjson import OrJSONSerializer
from stash.serializers.rapidjson import RapidJSONSerializer
from stash.serializers.serializer import Serializer
from stash.serializers.serializer import Serializer
from stash.serializers.simplejson import SimpleJSONSerializer
from stash.serializers.ujson import UltraJSONSerializer
from stash.storages.dbm_ import DbmStorage
from stash.storages.filesystem import FileSystemStorage
from stash.storages.leveldb import LeveldbStorage
from stash.storages.lmdb import LmdbStorage
from stash.storages.lsmdb import LsmDbStorage
from stash.storages.memory import MemoryStorage
from stash.storages.mongodb import MongoDbStorage
from stash.storages.null import NullStorage
from stash.storages.redis import RedisStorage
from stash.storages.storage import Storage

__name__: str = consts.PROJECT
__author__: str = "Masroor Ehsan"
__email__: str = "masroore@gmail.com"
__version__: str = ".".join(map(str, consts.VERSION))

__all__: List[str] = [
    "StashManager",
    # serializers
    "Serializer",
    "DefaultSerializer",
    "BSONSerializer",
    "CBORSerializer",
    "JSONSerializer",
    "MsgPackSerializer",
    "OrJSONSerializer",
    "RapidJSONSerializer",
    "SimpleJSONSerializer",
    "UltraJSONSerializer",
    # codecs
    "Codec",
    "PassthruCodec",
    "BrotliCodec",
    "BZip2Codec",
    "GZipCodec",
    "Lz4Codec",
    "LzfCodec",
    "LzmaCodec",
    "LzoCodec",
    "SnappyCodec",
    "ZlibCodec",
    "ZopfliCodec",
    "ZstdCodec",
    # storages
    "Storage",
    "DbmStorage",
    "FileSystemStorage",
    "LmdbStorage",
    "LeveldbStorage",
    "LsmDbStorage",
    "MemoryStorage",
    "NullStorage",
    "MongoDbStorage",
    "RedisStorage",
]



try:
    import lmdb
except ImportError:
    pass

from stash.options import StashOptions
from stash.storages.storage import Storage
from stash.utils.checksum import to_bytes


class LmdbStorage(Storage):
    def __init__(self, options: StashOptions):
        super().__init__(options)
        self.__cache_max_age = self.options.cache_max_age
        self._logger = self.options.logger
        self._env = lmdb.open(
            self.options.fs_cache_dir, max_dbs=1, map_size=self.options.lmdb_map_size
        )
        self._db = self._env.open_db(
            self.normalize_string(self.options.fs_cache_filename)
        )

    def exists(self, key: str) -> bool:
        key = key.encode("utf-8")
        with self._env.begin(db=self._db) as txn:
            for k, _ in txn.cursor():
                if k == key:
                    return True

        return False

    def purge(self, cutoff: int):
        pass

    @staticmethod
    def normalize_string(key: str) -> bytes:
        return to_bytes(key.strip())

    def clear(self):
        with self._env.begin(write=True, db=self._db) as txn:
            txn.drop(self._db, delete=False)

    def close(self):
        self._env.close()

    def write(self, key: str, content):
        key = self.normalize_string(key)
        with self._env.begin(write=True, db=self._db) as txn:
            txn.put(key, content)

    def read(self, key: str):
        key = self.normalize_string(key)
        with self._env.begin(db=self._db) as txn:
            return bytes(txn.get(key))

    def rm(self, key: str):
        key = self.normalize_string(key)
        with self._env.begin(write=True, db=self._db) as txn:
            txn.delete(key)



from stash.manager import StashManager
from stash.options import StashOptions


def _init_cache(storage, codec, options: StashOptions) -> StashManager:
    cache_man = StashManager(storage=storage, codec=codec, options=options)
    return cache_man


def _init_fs_cache(codec, options: StashOptions) -> StashManager:
    from .storages.filesystem import FileSystemStorage

    return _init_cache(
        storage=FileSystemStorage(options=options), codec=codec, options=options
    )


def get_fs_zlib_stash(options: StashOptions) -> StashManager:
    from .codecs.zlib import ZlibCodec

    return _init_fs_cache(ZlibCodec(), options=options)


def get_fs_brotli_stash(options: StashOptions) -> StashManager:
    from .codecs.brotli import BrotliCodec

    return _init_fs_cache(BrotliCodec(), options=options)


def get_fs_zstd_stash(options: StashOptions) -> StashManager:
    from .codecs.zstd import ZstdCodec

    return _init_fs_cache(ZstdCodec(), options=options)


def get_mongo_zlib_stash(options: StashOptions) -> StashManager:
    from .codecs.zlib import ZlibCodec
    from .storages.mongodb import MongoDbStorage

    storage = MongoDbStorage(options=options)
    return _init_cache(storage, ZlibCodec(), options=options)


def get_lmdb_zl_stash(options: StashOptions) -> StashManager:
    from .codecs.zlib import ZlibCodec
    from .storages.lmdb import LmdbStorage

    storage = LmdbStorage(options=options)
    return _init_cache(storage, ZlibCodec(), options=options)


def get_lmdb_zstd_stash(options: StashOptions) -> StashManager:
    from .codecs.zstd import ZstdCodec
    from .storages.lmdb import LmdbStorage

    storage = LmdbStorage(options=options)
    return _init_cache(storage, ZstdCodec(), options=options)


def get_fs_stash(options: StashOptions) -> StashManager:
    return _init_fs_cache(codec=None, options=options)


def get_null_stash() -> StashManager:
    from .storages.null import NullStorage
    from .codecs.passthru import PassthruCodec

    options = StashOptions()
    return _init_cache(
        storage=NullStorage(options=options), codec=PassthruCodec(), options=options
    )


def get_dbm_zstd_stash(options: StashOptions) -> StashManager:
    from .codecs.zstd import ZstdCodec
    from .storages.dbm_ import DbmStorage

    storage = DbmStorage(options=options)
    return _init_cache(storage, ZstdCodec(), options=options)


def get_dbm_zlib_stash(options: StashOptions) -> StashManager:
    from .codecs.zlib import ZlibCodec
    from .storages.dbm_ import DbmStorage

    storage = DbmStorage(options=options)
    return _init_cache(storage, ZlibCodec(), options=options)


def get_dbm_stash(options: StashOptions) -> StashManager:
    from .storages.dbm_ import DbmStorage
    from .codecs.passthru import PassthruCodec

    storage = DbmStorage(options=options)
    return _init_cache(storage, codec=PassthruCodec(), options=options)


def get_lsmdb_zstd_stash(options: StashOptions) -> StashManager:
    from .codecs.zstd import ZstdCodec
    from .storages.lsmdb import LsmDbStorage

    storage = LsmDbStorage(options=options)
    return _init_cache(storage, ZstdCodec(), options=options)


def get_lsmdb_zlib_stash(options: StashOptions) -> StashManager:
    from .codecs.zlib import ZlibCodec
    from .storages.lsmdb import LsmDbStorage

    storage = LsmDbStorage(options=options)
    return _init_cache(storage, ZlibCodec(), options=options)


def get_lsmdb_stash(options: StashOptions) -> StashManager:
    from .storages.lsmdb import LsmDbStorage
    from .codecs.passthru import PassthruCodec

    storage = LsmDbStorage(options=options)
    return _init_cache(storage, codec=PassthruCodec(), options=options)



import zlib
from stash.codecs.codec import Codec


class ZlibCodec(Codec):
    def encode(self, data):
        return zlib.compress(data)

    def decode(self, data):
        return zlib.decompress(data)



from stash.storages.storage import Storage


class NullStorage(Storage):
    def exists(self, key: str) -> bool:
        return False

    def purge(self, cutoff: int):
        pass

    def clear(self):
        pass

    def close(self):
        pass

    def write(self, key: str, content):
        pass

    def read(self, key: str):
        pass

    def rm(self, key: str):
        pass



from stash import consts, defaults

_default = {
    "algo": defaults.CACHE_DEFAULT_ALGORITHM,
    "cache_max_age": consts.SECONDS_IN_DAY,
    "cache_min_size": consts.SIZE_KB * 4,
    "fs_cache_filename": defaults.CACHE_DEFAULT_FILENAME,
    "fs_cache_dir": defaults.CACHE_DEFAULT_PATH,
    "fs_cache_dir_level": defaults.CACHE_DEFAULT_DIR_LEVEL,
    "fs_cache_segment_size": defaults.CACHE_SEGMENT_SIZE,
    "fs_cache_file_ext": "",
    "logger": None,
    "lmdb_map_size": consts.SIZE_MB * 4,
    "lmdb_filename": defaults.CACHE_DEFAULT_FILENAME,
    "dbm_filename": defaults.CACHE_DEFAULT_FILENAME + ".dbm",
    "lsmdb_filename": defaults.CACHE_DEFAULT_FILENAME + ".ldb",
    "redis_unix_socket_path": None,
    "redis_host": "localhost",
    "redis_port": 6379,
    "redis_db": 0,
    "redis_ssl": False,
    "redis_ssl_ca_certs": None,
}


class StashOptions(object):
    def __init__(self, data=None):
        if data is None:
            data = _default
        else:
            data |= _default
        super().__setattr__("data", {})
        self.data = data

    def __getattr__(self, name):
        if name in self.data:
            return self.data[name]
        return None

    def __setattr__(self, key, value):
        if key in self.data:
            self.data[key] = value
        else:
            super().__setattr__(key, value)



from stash.serializers.serializer import Serializer

try:
    from bson import decode, encode
except ImportError:
    pass


class BSONSerializer(Serializer):
    def deserialize(self, data):
        return decode(data)

    def serialize(self, data):
        return encode(data)



from stash.serializers.serializer import Serializer

try:
    from msgpack import loads, dumps
except ImportError:
    pass


class MsgPackSerializer(Serializer):
    def deserialize(self, data):
        return loads(data)

    def serialize(self, data):
        return dumps(data)



import hashlib

from stash.consts import *

try:
    import xxhash as _xxhash
except ImportError:
    _xxhash = None

try:
    import murmurhash.mrmr as _murmurhash
except ImportError:
    _murmurhash = None


def calcsum(payload: str, algo: str) -> str:
    if not algo:
        return payload

    algo = algo.strip().lower()
    if algo == CHECKSUM_ALGO_XXH32:
        return calc_xxh32(payload)
    elif algo == CHECKSUM_ALGO_XXH64:
        return calc_xxh64(payload)
    elif algo == CHECKSUM_ALGO_MD5:
        return calc_md5(payload)
    elif algo == CHECKSUM_ALGO_SHA1:
        return calc_sha1(payload)
    elif algo == CHECKSUM_ALGO_MURMUR:
        return calc_murmur(payload)


def to_bytes(data) -> bytes:
    if isinstance(data, str):
        data = data.encode("utf-8")
    return data


def calc_md5(data: str) -> str:
    return hashlib.md5(to_bytes(data)).hexdigest().lower()


def calc_sha1(data: str) -> str:
    return hashlib.sha1(to_bytes(data)).hexdigest().lower()


def calc_xxh32(data: str) -> str:
    return _xxhash.xxh32(data).hexdigest().lower()


def calc_xxh64(data: str) -> str:
    return _xxhash.xxh64(data).hexdigest().lower()


def calc_xxh3_64(data: str) -> str:
    return _xxhash.xxh3_64(data).hexdigest().lower()


def calc_xxh3_128(data: str) -> str:
    return _xxhash.xxh3_128(data).hexdigest().lower()


def calc_murmur(data: str) -> str:
    return _murmurhash.hash(data).lower()



from typing import Tuple

PROJECT: str = "stash"
VERSION: Tuple[int, int, int] = (0, 0, 5)

CHECKSUM_ALGO_MD5 = "md5"
CHECKSUM_ALGO_SHA1 = "sha1"
CHECKSUM_ALGO_XXH32 = "xxh32"
CHECKSUM_ALGO_XXH64 = "xxh64"
CHECKSUM_ALGO_MURMUR = "murmur"

SIZE_KB = 1024
SIZE_MB = SIZE_KB * 1024
SIZE_GB = SIZE_MB * 1024

SECONDS_IN_DAY = 86400
SECONDS_IN_HOUR = 3600
SECONDS_IN_MINUTE = 60



import os

from .consts import CHECKSUM_ALGO_XXH64, SECONDS_IN_MINUTE

CACHE_DEFAULT_DIRNAME = ".cache"
CACHE_DEFAULT_FILENAME = "cache"
CACHE_DEFAULT_PATH = os.path.abspath("./" + CACHE_DEFAULT_DIRNAME)
CACHE_DEFAULT_DIR_LEVEL = 2
CACHE_SEGMENT_SIZE = 2
CACHE_FILE_MIN_SIZE = 2 * 1024
CACHE_MAX_AGE = 30 * SECONDS_IN_MINUTE
CACHE_DEFAULT_ALGORITHM = CHECKSUM_ALGO_XXH64



from stash.codecs.codec import Codec

try:
    import lz4.frame
except ImportError:
    pass


class Lz4Codec(Codec):
    def encode(self, data):
        return lz4.frame.compress(data)

    def decode(self, data):
        return lz4.frame.decompress(data)



import errno
import os
import shutil
import time
from stat import S_ISREG

from stash.options import StashOptions
from stash.storages.storage import Storage
from stash.utils.nested_path import NestedPathBuilder, walk_files, safe_delete_file


class FileSystemStorage(Storage):
    def __init__(self, options: StashOptions):
        super().__init__(options)
        self.__nested_path_builder = NestedPathBuilder(
            self.options.fs_cache_dir,
            self.options.fs_cache_dir_level,
            self.options.fs_cache_segment_size,
        )

    def resolve_filepath(self, key: str) -> str:
        return self.__nested_path_builder.resolve_path_ext(
            key, self.options.fs_cache_file_ext
        )

    def exists(self, key: str) -> bool:
        fpath = self.resolve_filepath(key)
        bigbang = time.time() - self.options.cache_max_age
        try:
            st = os.stat(fpath)
            # check if the cache file exists
            exists = S_ISREG(st.st_mode) and st.st_size > self.options.cache_min_size
            if exists:
                # check cache freshness
                if st.st_mtime < bigbang:
                    # cache is stale. nuke it
                    if self.options.logger:
                        self.options.logger.debug(
                            "Deleting stale cache item: {} Last modified: {}".format(
                                os.path.basename(fpath), time.ctime(st.st_mtime)
                            )
                        )
                    safe_delete_file(fpath)
                    return False
                # cache file exists and is fresh & valid
                return True
            # cache file is non-existent or invalid
            return False
        except (IOError, OSError) as e:
            if e.errno != errno.ENOENT:
                if self.options.logger:
                    self.options.logger.error(e)
                    import traceback

                    self.options.logger.error(traceback.format_exc())

        return False

    def purge(self, cutoff: int):
        files = walk_files(
            self.options.fs_cache_dir, "*" + self.options.fs_cache_file_ext
        )
        bigbang = time.time() - cutoff
        for fname in files:
            try:
                st = os.stat(fname)
                if S_ISREG(st.st_mode):
                    if st.st_mtime < bigbang:
                        safe_delete_file(fname)
            except (IOError, OSError) as e:
                if e.errno != errno.ENOENT:
                    if self.options.logger:
                        self.options.logger.error(e)
                        import traceback

                        self.options.logger.error(traceback.format_exc())

    def clear(self):
        try:
            shutil.rmtree(self.options.fs_cache_dir, ignore_errors=True)
        except (IOError, OSError) as e:
            if e.errno != errno.ENOENT:
                if self.options.logger:
                    self.options.logger.error(e)
                    import traceback

                    self.options.logger.error(traceback.format_exc())

    def close(self):
        pass

    def write(self, key: str, content):
        with open(self.resolve_filepath(key), "wb") as f:
            f.write(content)

    def read(self, key: str):
        with open(self.resolve_filepath(key), "rb") as f:
            return f.read()

    def rm(self, key: str):
        fpath = self.resolve_filepath(key)
        if os.path.isfile(fpath):
            os.unlink(fpath)



from __future__ import absolute_import

from stash import defaults
from stash.codecs.codec import Codec
from stash.options import StashOptions
from stash.serializers.default import DefaultSerializer
from stash.serializers.serializer import Serializer
from stash.storages.storage import Storage
from stash.utils.checksum import calcsum, to_bytes


class StashManager(object):
    def __init__(
        self,
        storage: Storage,
        codec: Codec,
        options: StashOptions,
        serializer: Serializer = DefaultSerializer(),
    ):
        self.__storage = storage
        self.__codec = codec
        self.__options = options
        self.__serializer = serializer

    def __get_cache_key(self, data):
        return calcsum(data, self.__options.algo)

    def __encode(self, data):
        data = to_bytes(data)
        return self.__codec.encode(data) if self.__codec else data

    def __decode(self, data):
        return self.__codec.decode(data) if self.__codec else data

    def exists(self, key: str) -> bool:
        return self.__storage.exists(self.__get_cache_key(key))

    def purge(self, cutoff: int = defaults.CACHE_MAX_AGE):
        return self.__storage.purge(cutoff)

    def clear(self):
        return self.__storage.clear()

    def close(self):
        return self.__storage.close()

    def write(self, key: str, content):
        data = self.__encode(content)
        return self.__storage.write(self.__get_cache_key(key), data)

    def read(self, key: str):
        data = self.__storage.read(self.__get_cache_key(key))
        if data:
            return self.__decode(data)
        return None

    def rm(self, key: str):
        self.__storage.rm(self.__get_cache_key(key))



from stash.codecs.codec import Codec

try:
    import gzip
except ImportError:
    pass


class GZipCodec(Codec):
    def encode(self, data):
        return gzip.compress(data)

    def decode(self, data):
        return gzip.decompress(data)



try:
    from redis import Redis
except ImportError:
    pass

from stash.options import StashOptions
from stash.storages.storage import Storage


class RedisStorage(Storage):
    def __init__(self, options: StashOptions):
        super().__init__(options)
        if self.options.redis_unix_socket_path:
            self.client = Redis(unix_socket_path=self.options.redis_unix_socket_path)
        else:
            self.client = Redis(
                host=self.options.redis_host,
                port=self.options.redis_port,
                db=self.options.redis_db,
                ssl=self.options.redis_ssl,
                ssl_ca_certs=self.options.redis_ssl_ca_certs,
            )

    def exists(self, key: str) -> bool:
        return self.client.exists(key) > 0

    def purge(self, cutoff: int):
        pass

    def clear(self):
        self.client.flushdb()

    def close(self):
        self.client.close()

    def write(self, key: str, content):
        # item = {"data": Binary(content), "timestamp": datetime.utcnow()}
        self.client.set(key, content)

    def read(self, key: str):
        return self.client.get(key)

    def rm(self, key: str):
        self.client.delete(key)



from stash.codecs.codec import Codec

try:
    from zopfli.zlib import compress
    from zlib import decompress
except ImportError:
    pass


class ZopfliCodec(Codec):
    def encode(self, data):
        return compress(data)

    def decode(self, data):
        return decompress(data)



from stash.codecs.codec import Codec

try:
    import snappy
except ImportError:
    pass


class SnappyCodec(Codec):
    def encode(self, data):
        return snappy.compress(data)

    def decode(self, data):
        return snappy.uncompress(data)



from stash.codecs.codec import Codec

try:
    import bz2
except ImportError:
    pass


class BZip2Codec(Codec):
    def encode(self, data):
        return bz2.compress(data)

    def decode(self, data):
        return bz2.decompress(data)



from datetime import datetime

try:
    from pymongo import MongoClient
    from bson import Binary
except ImportError:
    pass

from stash.options import StashOptions
from stash.storages.storage import Storage


class MongoDbStorage(Storage):
    def __init__(self, options: StashOptions):
        super().__init__(options)
        self.client = MongoClient(
            self.options.mongo_dsn  # "mongodb://localhost:27017/"
        )
        self.db = self.client.cache_storage
        self.cache_items = self.db.cache_items
        self.cache_items.create_index(
            "timestamp", expireAfterSeconds=self.options.cache_max_age
        )

    def exists(self, key: str) -> bool:
        return self.cache_items.find_one({"_id": key}) is not None

    def purge(self, cutoff: int):
        pass

    def clear(self):
        self.cache_items.drop()

    def close(self):
        self.client.close()

    def write(self, key: str, content):
        item = {"data": Binary(content), "timestamp": datetime.utcnow()}
        self.cache_items.update_one({"_id": key}, {"$set": item}, upsert=True)

    def read(self, key: str):
        item = self.cache_items.find_one({"_id": key})
        if item:
            return item["data"]
        return None

    def rm(self, key: str):
        pass



from stash.codecs.codec import Codec

try:
    import zstd
except ImportError:
    pass


class ZstdCodec(Codec):
    def encode(self, data, level=10):
        return zstd.compress(data, level)

    def decode(self, data):
        return zstd.decompress(data)



import os
from time import time
import dbm
from stash.options import StashOptions
from stash.storages.storage import Storage


class DbmStorage(Storage):
    def __init__(self, options: StashOptions):
        super().__init__(options)
        dbpath = os.path.join(self.options.fs_cache_dir, options.dbm_filename)
        self.__db = dbm.open(dbpath, "c")

    def _data_key(self, key: str) -> str:
        return f"{key.strip()}^@d"

    def _meta_key(self, key: str) -> str:
        return f"{key.strip()}^@m"

    def exists(self, key: str) -> bool:
        # if self._meta_key(key) not in self.__db.keys(): return False
        return self._data_key(key) in self.__db.keys()

    def purge(self, cutoff: int):
        pass

    def clear(self):
        self.__db.clear()

    def close(self):
        self.__db.close()

    def write(self, key: str, content):
        self.__db[self._data_key(key)] = content
        self.__db[self._meta_key(key)] = str(time())

    def read(self, key: str):
        return self.__db.get(self._data_key(key))

    def rm(self, key: str):
        self.__db.pop(key)



import errno
import fnmatch
import os


def split_segments(s: str, level: int, segment_size: int) -> list:
    segments = [
        s[i : i + segment_size]
        for i in range(1, 1 + (segment_size * level), segment_size)
    ]
    segments.append(s[1 + (segment_size * level) :])
    return segments


def mkdir_p(path: str):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def safe_delete_file(fname: str):
    try:
        os.unlink(fname)
    except:
        pass


def walk_files(folder: str, pattern: str):
    matches = []
    for root, dirnames, filenames in os.walk(folder):
        for filename in fnmatch.filter(filenames, pattern):
            matches.append(os.path.join(root, filename))
    return matches


class NestedPathBuilder(object):
    def __init__(self, base_dir: str, dir_level: int, segment_size: int):
        self.__base_dir = base_dir
        self.dir_level = dir_level
        self.segment_size = segment_size

    def __resolve_dir_path(self, key: str) -> tuple:
        if not os.path.exists(self.__base_dir):
            mkdir_p(self.__base_dir)
        path = os.sep.join(split_segments(key, self.dir_level, self.segment_size))
        path = os.path.abspath(os.path.join(self.__base_dir, path))
        dirname = os.path.dirname(path)
        mkdir_p(dirname)
        return dirname, path

    def resolve_path(self, key: str) -> str:
        _, path = self.__resolve_dir_path(key)
        return path

    def resolve_dir(self, key: str) -> str:
        dirname, _ = self.__resolve_dir_path(key)
        return dirname

    def resolve_path_ext(self, key: str, suffix: str = None) -> str:
        # return os.path.join(self.resolve_dir(key), key + suffix if suffix else key)
        path = self.resolve_path(key)
        return path + suffix if suffix else path



from stash.codecs.codec import Codec

try:
    import lzo
except ImportError:
    pass


class LzoCodec(Codec):
    def encode(self, data):
        return lzo.compress(data)

    def decode(self, data):
        return lzo.decompress(data)
